/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 1                                                                */
/*                                                                       */
/* Coded by:  L.W. Finger                                                */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            finger@gl.ciw.edu                                          */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/* martin@oc2.oc.chemie.th-darmstadt.de                                  */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

#include "drawxtl.h"

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#ifdef _MAC_              /* special includes for Mac version */
#include <console.h>
#include <sioux.h>
#endif

/* forward references for routines */

void add_to_list();
void add_vert();
void add_vert_nc();
void analyze_bonds();
void axeqb();
void Blank_Strip();
void build_box_contents();
void build_poly_list();
void Calc_Rot();
void check_atom_label();
void check_color();
void convert_ellipsoid();
void crystal_compute();
float determinant();
float dist();
void do_drawing();
void draw_cell();
void expand_atom();
int eigen();
void find_all_in_box();
void generate_bonds();
void generate_ellipsoids();
void generate_planes();
void generate_poly();
void generate_spheres();
void get_atom_id();
void get_input();
void get_label();
void import_cif();
void import_gsas();
void import_schakal();
void import_shelx();
void import_fdat();
void Locate_Triple();
void make_bmat();
float P_to_C();
void Output_Spheres();
void plot_vrml_poly();
void print_sym();
void read_inp();
void symop();
void Token_Strip();
void usage();

/* global variables for lattice constants and symmetry operators */

char spg[3][4];                 /* space group code */
int ss[24][3][3];               /* rotation operators for symmetry */
float lat_pos[4][3];            /* coordinates of lattice points */
float ts[24][3];                /* translation operators */
int sym_op_no[192];             /* saved symmetry operation number */
float cell_xyz[192][3];         /* storage for generated equiv. positions */
int sys;                        /* crystal system indicator */
int nbr;                        /* Bravais lattice indicator */
int ns;                         /* center of symmetry indicator */
int ng;                         /* number of rotation operations to specify group */
int nlat;                       /* number of lattice positions */
float lat_con[6];               /* lattice constants */
float ginv[3][3];               /* direct space metric tensor */
float b_inv_mat[3][3];          /* inverse metric tensor */
float b_mat[3][3];              /* metric tensor */
float rec_lat_con[3];           /* reciprocal lengths */

/* global variables for file names and file variables */

char  savedpath[100];           /* path at start of program */
char  File_Name[100];           /* Modified Input File Name */
FILE  *fpin,*flout;
FILE  *fpoutv;
FILE  *fpoutp;
char  *Str_File_Inp;            /* Input File Name */
char  *Str_File;
char  *output;

/* global variables for atom lists */

int   natom;                    /* number of different types of atoms */
int   ncell;                    /* number of atoms in asymmetric unit */
int   atom_no[MAX_VERTS];       /* master list atom and symmetry number */
float xypos[3*MAX_VERTS];       /* master list fractional coordinates */
float *o_vert;                  /* crystal coordinates of saved vertices */
float *s_vert;                  /* saved coordinates of vertices */
int   *vert_sym_no;             /* saved symmetry operator number */
int   *poly_list;               /* storage for polygon corner pointers */
int   nvert;                    /* number of vertices in list */
int   NvertM;                    /* number of vertices in master atom list */
float atom_xyz[MAX_ATOM][3];    /* atomic coordinates */
char  atom_l[MAX_ATOM][4];      /* atom labels */
int   atom_n[MAX_ATOM];         /* input atom number */
int   sv_atom_n[MAX_ATOM];      /* saved input atom number */

/* global variables for spheres and ellipsoids */

int   nsphere;                  /* number of different types of spheres */
int   n_ellips;                 /* no. of aniso. atoms in asymmetric unit */
int   do_ellipsoids;            /* non-zero if ellipsoids to be drawn */
int   El_Cutout;                /* True if ellipsoids to have an octant removed */
float Sphere_Mult;              /* Multiplier for Sphere Size */
float SpMult;                   /* initialization for Sphere_Mult */
float Ellipsoid_Prob;           /* Fraction of electron dens enclosed by ellipsoid */
float Ellipsoid_Scale;          /* Scaling needed to get probability */
float Phong_Value;              /* reflectivity value for Phong highlighting */
float Phong_Size;               /* Phong highlighting size */
char  sphere_l[MAX_SPHERE][4];  /* sphere label */
float sphere_size[MAX_SPHERE];  /* sphere size */
char  sphere_col_v[MAX_SPHERE][30];/* sphere colors */
char  sphere_col_p[MAX_SPHERE][30];/* sphere colors */
char  ellips_l[MAX_EL][4];      /* label for anisotropic atom */
int   ellips_n[MAX_EL];         /* atom number */
int   ell_type[MAX_EL];         /* form of ellipsoid input */
float ellips[MAX_EL][6];        /* thermal ellipsoid coefficients */
float ellips_RMS[MAX_EL][3];    /* RMS amplitudes */
float ellips_EV[MAX_EL][3][3];  /* Eigenvectors */
char  ellips_col_v[MAX_EL][30]; /* ellipsoid colors for VRML */
char  ellips_col_p[MAX_EL][30]; /* ellipsoid colors for POV */
char  Cutout_color[30];         /* color of cutout section */

/* global variables for polyhedra, planes and bonds */

int   npoly;                    /* number of different types of polyhedra */
int   nplane;                   /* number of different types of planes */
int   nbond;                    /* number of different types of bonds */
int   draw_list;                /* number of items in polygon draw list */
int   *vertex_list;             /* storage for vertices about a cation */
int   numb_list;                /* number in polygon list */
int   domolcomp;                /* non-zero if molecule completion requested */
float polyfudge;                /* should be close to zero to allow only regular polygonal faces */
float Bond_Mult;                /* Multiplier for Bond Size */
float BndMult;                  /* Bond multiplier initialization */
float mol_d;                    /* intramolecular distance for molecular and polyhedral completion */
char  poly_l[MAX_POLY][4];      /* polyhedral center label */
float poly_size[MAX_POLY];      /* size of polyhedron */
char  poly_col_v[MAX_POLY][30]; /*polyhedron colors for VRML */
char  poly_col_p[MAX_POLY][30]; /*polyhedron colors for POV */
char  plane_col_v[MAX_PLANE][30];/*plane colors for VRML */
char  plane_col_p[MAX_PLANE][30];/*plane colors for POV */
char  plane_l[MAX_PLANE][4];    /* plane center label */
float plane_size[MAX_PLANE];    /* size of plane coordination */
char  bond_l1[MAX_BOND][4];     /* label for one end of bond */
char  bond_l2[MAX_BOND][4];     /* label for other end */
float bond_size[MAX_BOND];      /* bond diameter */
float bond_min[MAX_BOND];       /* minimum bond distance */
float bond_max[MAX_BOND];       /* maximum bond distance */
char  col_bond_v[MAX_BOND][30]; /* color and properties of bonds for VRML */
char  col_bond_p[MAX_BOND][30]; /* color and properties of bonds for POV */
                         
/* global variables that control size and orientation of output object */

float  DepthCue;                /* scale for Z-dependent thickness of edges*/
float X_Boxlim;                 /* X box limit read from command line */
float Y_Boxlim;                 /* Y box limit read from command line */
float Z_Boxlim;                 /* Z box limit read from command line */
float X_Origin;                 /* X origin read from command line */
float Y_Origin;                 /* Y origin read from command line */
float Z_Origin;                 /* Z origin read from command line */
float boxlim[3];                /* half limits of plotting box */
float origin[3];                /* position of plotting origin */
float G_Rot[3][3];              /* Grand rotation matrix - transforms from Cartesian to picture */
float lookat_v1[3],lookat_v2[3];/* "lookat" vectors " */
float POV_Max[3];               /* Maximum limits of POV after rotation */
float POV_Min[3];               /* Minimum limits of POV after rotation */
int   boxflag;                  /* Non-zero if 'bounds' command given */
int   packflag;                 /* Non-zero if 'pack' command given */
int   docell;                   /* non-zero if unit-cell edges to be drawn */
int   Display_axes;             /* non-zero if axial triple to be drawn */
int   Color_Warning;            /* Warning flag for non-standard color */
float Magnification;            /* Magnification factor for image */
float Scale;                    /* Scale of diagram */
float offset[3];                /* offset for vector triple */
float Text_Size;                /* Size to make axis label text */
int   Options;                  /* Place to save command-line options flags */
char  col_cell_v[30];           /* color and properties of framebars */
char  col_edge_v[30];           /* color and properties of polyhedron edges */
char  col_cell_p[30];           /* color and properties of framebars */
char  col_edge_p[30];           /* color and properties of polyhedron edges */
float rad_cell;                 /* radius of unit cell framebars */
int   edges;                    /* draw thin black lines around the edges of polyhedra*/
float rad_edge;                 /* radius of cylinders on polyhedron edges */
float xrot,yrot,zrot;           /* view rotation angles */
float cryst_lim[6];             /* limits of view box in cell coords */
int   Unit_Cell;                /* non-zero if unit cell to be drawn */
int   no_comment;               /* non-zero to inhibit comment lines in VRML */
int   M_cameras;                /* non-zero to inhibit multiple cameras */

#define GEOMedit_list char
#define GEOMobj int

/* ************************************************************** */
/* ************************************************************** */

int main(int argc,char *argv[])

{
    int i;

#ifdef _MAC_
    argc = ccommand(&argv);            /* get command line for Macintosh */
    SIOUXSettings.showstatusline=TRUE; /* show program status in window */
#endif

/* Initialize */
    
    X_Origin = Y_Origin = Z_Origin = 0.5;
    X_Boxlim = Y_Boxlim = Z_Boxlim = 20.;
    xrot = yrot = zrot = 0.0;
    boxflag = packflag = domolcomp = 0;
    do_ellipsoids = 0;
    Ellipsoid_Prob = 0.5;
    Unit_Cell = M_cameras = no_comment = edges = 1;
    Display_axes = El_Cutout = 0;
    strcpy(Cutout_color,"Gray20");
    SpMult = BndMult = Magnification = 1.;
    strcpy(col_cell_v,"0 0 0");  /* preset unit-cell colors to Black */
    strcpy(col_cell_p,"Black");
    Ellipsoid_Scale = P_to_C(0.5);
    Options = 0;
    polyfudge = (float)0.1;
    Phong_Value = (float)0.2;
    Phong_Size = 1.0;
    DepthCue = 0.;

/* Parse arguments */

    if (argc <2) usage();

    for (i = 1; i < argc; i++) {
	if (!strcmp(argv[i], "-n"))
	    Unit_Cell = 0;
	else if (!strcmp(argv[i], "-h")) usage();
	else if (!strcmp(argv[i], "-?")) usage();
	else if (!strcmp(argv[i], "-o")){
            X_Origin = (float)atof(argv[++i]);
            Y_Origin = (float)atof(argv[++i]);
            Z_Origin = (float)atof(argv[++i]);
            Options = Options | O_OPT;
	  }
        else if (!strcmp(argv[i], "-C")){
            M_cameras = 0;
          }
        else if (!strcmp(argv[i], "-c")){
            no_comment = 0;
          }
	else if (!strcmp(argv[i], "-b")){
	    X_Boxlim = (float)atof(argv[++i]);
	    Y_Boxlim = (float)atof(argv[++i]);
	    Z_Boxlim = (float)atof(argv[++i]);
	    boxflag=1;
            Options = Options | B_OPT;
	  }
       else if (!strcmp(argv[i], "-p")){
	    cryst_lim[0] = (float)atof(argv[++i]);
	    cryst_lim[3] = (float)atof(argv[++i]);
	    cryst_lim[1] = (float)atof(argv[++i]);
	    cryst_lim[4] = (float)atof(argv[++i]);
	    cryst_lim[2] = (float)atof(argv[++i]);
	    cryst_lim[5] = (float)atof(argv[++i]);
            Options = Options | P_OPT;
            packflag = 1;
	  }
	else if (!strcmp(argv[i], "-v")){
	    xrot = (float)atof(argv[++i]);
	    yrot = (float)atof(argv[++i]);
	    zrot = (float)atof(argv[++i]);
            Options = Options | V_OPT;
	  }
        else if (!strcmp(argv[i], "-m")){
            Magnification = (float)atof(argv[++i]);
            Options = Options | M_OPT;
	  }
        else if (!strcmp(argv[i], "-u")){
            Display_axes = 1;
	  }
	else 
          Str_File=argv[i];
      }
      if (packflag != 0) {
        X_Boxlim = Y_Boxlim = Z_Boxlim = 500.0;
      }
      crystal_compute();      /* convert to POV or VRML */
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      return 0;
} /* end main for POV or VRML versions */
 
/* ************************************************************** */
/* ************************************************************** */

void add_vert(vert,no,check)

float vert[3];                  /* input coordinates of vertices */
int no;                         /* symmetry operator number */
int check;                      /* non-zero to skip in box check */

/* procedure to add the vertex in 'vert' to the current list if the
   vertex is contained within the limits of the box, within the limits
   of a polyhedral center, or needed to complete a molecule */

{
  float c_vert[3];                    /* vertex in cartesian coords */
  float r_vert[3];                    /* vertex in rotated coordinates */
  int i,j;                            /* loop variables */
  float temp;                         /* temporary storage */
  int outside;                        /* temporary for molecular completion */
  int number;                         /* decoded atom number */
  int poly_center;                    /* true if this atom is center of a polyhedron */
  int consider_extra;
  float d;
  
  d = 0.0;
  for (i=0;i<3;++i) {  /* convert vertex coordinates to Cartesian */
    c_vert[i] = 0.0;
    for (j=0;j<3;++j)                             
      c_vert[i] += b_mat[i][j] * (vert[j] - origin[j]);
  }
  number = no/100;
  poly_center = (atom_n[number] > 0 && atom_n[number] < 10000);
  consider_extra = (domolcomp != 0 || (npoly > 1 && !poly_center));
  outside = 0;
  if (check == 0){   
    for (i=0;i<=2;i++){  /* check packing range */
      if ((vert[i] < cryst_lim[i]) || (vert[i] > cryst_lim[i+3] )) 
        outside=1;
    }
    if ((outside == 0) && (fabs(c_vert[0]) > boxlim[0]) || 
      (fabs(c_vert[1]) > boxlim[1]) || (fabs(c_vert[2]) > boxlim[2]))
        outside = 1;  /* check box limits */
    if (outside != 0 && !consider_extra) return;  
  }
  d = 0.0;
  if (outside == 1) {     /* molecule or polyhedral completion is in progress */
    float dmax;
    int polyno;           /* number of the polyhedron */
    dmax = mol_d*mol_d;
    for (i=1;i<NvertM;i++) {
      d = 0.0;
      for (j=0;j<3;j++){
        temp = (s_vert[3*i+j] - c_vert[j]);
        d += temp*temp;
      }
      j = atom_no[i]/100;
      poly_center = (atom_n[j] > 0 && atom_n[j] < 10000);
      if (poly_center) {
        polyno = atom_n[j];
        dmax = poly_size[polyno]*poly_size[polyno];
      }
      if ((domolcomp != 0 || poly_center) && (d <= dmax && d > 0.02)){
        outside = 2;
        break;
      }
    }
  }
  if (outside == 1)return;
  i = (int) (no/100);
  if (number == i) {
    for (j=1;j<NvertM;++j){
      if (fabs(s_vert[3*j]  - c_vert[0]) < 0.01 &&
         fabs(s_vert[3*j+1] - c_vert[1]) < 0.01 &&
         fabs(s_vert[3*j+2] - c_vert[2]) < 0.01 &&
         atom_l[i][0]==atom_l[atom_no[j]/100][0] &&
         atom_l[i][1]==atom_l[atom_no[j]/100][1] &&
         atom_l[i][2]==atom_l[atom_no[j]/100][2] &&
         atom_l[i][3]==atom_l[atom_no[j]/100][3] )
           return;    /* eliminate duplicates */
    }
  }
  for (i=0;i<=2;++i){   /* add atom to Master List */
    xypos[3*NvertM + i]  = vert[i];
    s_vert[3*NvertM + i] = c_vert[i];
    if (offset[i] > c_vert[i]) offset[i] = c_vert[i];
  }
  atom_no[NvertM++] = no;
  if (NvertM > MAX_VERTS) {
    if (domolcomp != 0) {
      fprintf(stdout,"Too many vertices for dimensions:\n");
      fprintf(stdout," Did you use a 'molcomp' command incorrectly?\n");
      fprintf(stdout," If not, increase MAX_VERTS.\n");
    } else {
      fprintf(stdout,"Too many vertices for dimensions - increase MAX_VERTS.\n");
    }
#ifdef _MAC_
    printf("Use Apple Q to close this window.\n");
#endif
    exit(0);                       
  }
  for(i=0;i<=2;++i){       /* calculate position of point after POV rotation */
    r_vert[i] = 0.0;
    for(j=0;j<=2;++j)
      r_vert[i] += G_Rot[j][i] * c_vert[j];
  }
  for(i=0;i<=2;++i){       /* Update Min and Max of output */
    if (POV_Max[i] < r_vert[i] ) POV_Max[i] = r_vert[i];
    if (POV_Min[i] > r_vert[i] ) POV_Min[i] = r_vert[i];
  }                            
}

/* ************************************************************** */
/* ************************************************************** */

void add_vert_nc(vert)

float vert[3];                  /* input coordinates of vertices */

/* procedure to add the vertices in 'vert' to the current working list
   (NOT Master List ) without checking */

{

  int i,j;                            /* loop variables */

  for (i=0;i<=2;++i) {  /* convert vertex coordinates to Cartesian */
    s_vert[3*nvert+i] = 0.0;
    for (j=0;j<=2;++j)                             
      s_vert[3*nvert+i] += b_mat[i][j] * (vert[j] - origin[j]);
  }
  nvert++;
}

/* ************************************************************** */
/* ************************************************************** */

void add_to_list(xp,no)

/* add atom position to unit-cell list if unique */

float xp[3];
int no;                 /* symmetry operator number */

{
  int i,ind;

  for (i=0;i<=2;++i) { /* get coordinates between 0 and 1 */
    if (xp[i] >= 1.0) xp[i] = xp[i] - (float)1.0;
    if (xp[i] < 0.0)  xp[i] = xp[i] + (float)1.0;
  }
  ind = 0;
  for (i=0;i<ncell;++i) {
    if ((fabs(cell_xyz[i][0] - xp[0]) <= 0.00002) &&
        (fabs(cell_xyz[i][1] - xp[1]) <= 0.00002) &&
        (fabs(cell_xyz[i][2] - xp[2]) <= 0.00002)) ind = 1; /*set if same*/
  }
  if (ind == 0) {  /* true for new position */
    for (i=0;i<=2;++i)  /*copy to list */
      cell_xyz[ncell][i] = xp[i];
    sym_op_no[ncell++] = no;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void analyze_bonds()

/* routine to print bond distances up to a maximum of 3.5 Angstrom */

{
  int i,j,k,l,nvert1,*itype;
  float *d,dmin;

  if ((itype = (int *) malloc((unsigned) (MAX_VERTS * sizeof(int)))) == NULL)
    return;
  fprintf(flout,"\n\n Bond Distance Analysis\n");
/* Find all atoms of this type in master list */
  nvert = nvert1 = 1;
  for(j=0;j<natom;++j) {
    l = nvert;
    find_all_in_box(j);
    for (k = l;k<nvert;++k)
      itype[k] = j;          /* itype[k] has atom number j */
  }
  if ((d = (float *) malloc((unsigned) ((nvert+10) * sizeof (float)))) == NULL) {
    free(itype);
    printf("Unable to get bond distance allocation\n");
    return; /* get space for distances */
  }
  for(i=0;i<natom;++i) { /* loop through atom types */
/* add the base entry to the beginning of the atom list */
    for (j=0;j<3;++j) {  /* convert vertex coordinates to Cartesian */
      s_vert[j] = 0.0;
      for (k=0;k<=2;++k)
        s_vert[j] += b_mat[j][k] * (atom_xyz[i][k]-origin[k]);
    }
    for (j=1;j<nvert;++j)  /* calculate distances */
      d[j] = dist(0,j);
    fprintf(flout,"\nDistances from %c%c%c%c%3d at %8.5f%8.5f%8.5f    to\n\n"
      ,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3],sv_atom_n[i]
      ,atom_xyz[i][0],atom_xyz[i][1],atom_xyz[i][2]);
    for (k=nvert1;k<nvert;++k) {
      dmin = (float)999999.9;
      for (j=nvert1;j<nvert;++j) {   /* loop through distances */
        if (d[j] < dmin) {
          l = j;
          dmin = d[j];
        }
      }
      if ( dmin == 0.0 && itype[l] == i) {
        d[l] = (float)999999.9;
      } else {
        if ( dmin < 3.5 ) {
          fprintf(flout," %c%c%c%c%3d at %9.5f%9.5f%9.5f  %8.3f\n"
           ,atom_l[itype[l]][0],atom_l[itype[l]][1],atom_l[itype[l]][2]
           ,atom_l[itype[l]][3],sv_atom_n[itype[l]],o_vert[3*l],o_vert[3*l+1]
           ,o_vert[3*l+2],dmin);
          d[l] = (float)999999.9;
        } else break;
      }
    }
  }
  free(d);
  free(itype);
  fprintf(flout,"\n\n");
}

/* ************************************************************** */
/* ************************************************************** */

void axeqb(a1,x,b)

/* solve matrix equation ax = b for x, where a is 3x3 and x and b are
   vectors */

double a1[3][3],x[3],b[3];

{
  int i,j,k;
  double a[3][3],det,t[3][3];

  det = a1[0][0] * (a1[1][1] * a1[2][2] - a1[1][2] * a1[2][1])
     - a1[1][0] * (a1[0][1] * a1[2][2] - a1[0][2] * a1[2][1])
     + a1[2][0] * (a1[0][1] * a1[1][2] - a1[0][2] * a1[1][1]);
  if (fabs(det) < 1.0e-6) det = 1.0e-6;
  a[0][0] =  (a1[1][1] * a1[2][2] - a1[1][2] * a1[2][1])/det;
  a[1][0] = -(a1[1][0] * a1[2][2] - a1[1][2] * a1[2][0])/det;
  a[2][0] =  (a1[1][0] * a1[2][1] - a1[1][1] * a1[2][0])/det;
  a[0][1] = -(a1[0][1] * a1[2][2] - a1[0][2] * a1[2][1])/det;
  a[1][1] =  (a1[0][0] * a1[2][2] - a1[0][2] * a1[2][0])/det;
  a[2][1] = -(a1[0][0] * a1[2][1] - a1[0][1] * a1[2][0])/det;
  a[0][2] =  (a1[0][1] * a1[1][2] - a1[0][2] * a1[1][1])/det;
  a[1][2] = -(a1[0][0] * a1[1][2] - a1[0][2] * a1[1][0])/det;
  a[2][2] =  (a1[0][0] * a1[1][1] - a1[0][1] * a1[1][0])/det;
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++){
      t[j][i] = 0.0;
      for (k=0;k<3;k++) {
	t[j][i] += a[k][i] * a1[j][k];
      }
    }
  }
  for(i=0;i<3;i++) {
    x[i] = 0.0;
    for (j=0;j<3;j++)
      x[i] += a[j][i] * b[j];
  }
}

/* ************************************************************** */
/* ************************************************************** */

void Blank_Strip(input)
char input[100];

/* routine to suppress any leading spaces in the input string */

{
  int i,j;

  for(i=0;i<100;i++)
    if (input[i] != ' ')break;
  for(j=0;j<100;j++) {
    input[j] = input[i];
    if (input[i++] == '\0')
      break;
  }
}

/* ************************************************************** */
/* ************************************************************** */

void build_box_contents()

/* routine to locate all atoms in the display parallelopiped plus
   any atoms needed to complete molecules or polyhedra */

{
  float vert[3];              /* storage for center */
  int ix[3];                  /* array for lattice offset */
  int j,k,saved_nvert;
  int atoms,loop;

  loop = 0;
  while (1) {             /* infinite loop here */
    saved_nvert = NvertM;  /* save number currently in list */
    for (atoms=0;atoms<natom;atoms++) { /* loop through atoms */
      expand_atom(atoms);     /* generate all atoms of this flavor in unit cell */
      for (j=0;j<ncell;++j) { /*atoms in cell*/
        for (k=0;k<=2;++k) 
           vert[k] = cell_xyz[j][k];
        add_vert(vert,100*atoms+sym_op_no[j],0);  /* add point to cell if inside box*/
      }
    }
    if ( saved_nvert == NvertM) break;
  }
  while (1) {             /* infinite loop here */
    saved_nvert = NvertM;  /* save number currently in list */
    for (atoms=0;atoms<natom;atoms++) { /* loop through atoms */
      expand_atom(atoms);     /* generate all atoms of this flavor in unit cell */
      for (j=0;j<ncell;++j) { /*atoms in cell*/
        for (k=0;k<=2;++k) 
           vert[k] = cell_xyz[j][k];
        add_vert(vert,100*atoms+sym_op_no[j],0);  /* add point to cell if inside box*/
      }
      for (ix[0] = (int) cryst_lim[0]-1;ix[0]<=(int) cryst_lim[3]+1;++ix[0]) { /* step through cells parallel a */
        for (ix[1] = (int) cryst_lim[1]-1;ix[1]<=(int) cryst_lim[4]+1;++ix[1]) { /* parallel b */
          for (ix[2] = (int) cryst_lim[2]-1;ix[2]<=(int) cryst_lim[5]+1;++ix[2]) { /* parallel c */
            if (abs(ix[0]) + abs(ix[1]) + abs(ix[2]) != 0) {
              for (j=0;j<ncell;++j) { /*atoms in cell*/
                for (k=0;k<=2;++k) 
                  vert[k] = ix[k] + cell_xyz[j][k];
                add_vert(vert,100*atoms+sym_op_no[j],0);  /* add point to cell if inside box*/
              }
            }
          }
        }
      }
    }
    if ((domolcomp == 0 && npoly == 1) || saved_nvert == NvertM) return;
    if (loop == 0)printf("\n");
    loop++;
    printf("Recursive Atom Addition, Loop No. %3d, %4d Atoms in List.\n",loop,NvertM-1);       
  }
}   /* end of build_box_contents */

/* ************************************************************** */
/* ************************************************************** */

void Calc_Rot(v1,v2)
float v1[3],v2[3];

/* routine to calculate rotation axes needed to put v1 parallel to Z
   and the projection of v2 parallel to Y */

{
  float u1[3],u2[3],u3[3],U[3][3],temp;
  int i,j;

  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      U[i][j] = 0.0;    /* clear U */
  temp = 0.0;
  for (i=0;i<3;i++) {  /* convert v1 to Cartesian */
    u1[i] = 0.0;
    for (j=0;j<3;j++)
      u1[i] += b_mat[i][j] * v1[j];
    temp += u1[i] * u1[i];
  }
  if (temp <= 0.0) {
    u1[2] = 1.0;   /* if null input, make u1 = 0,0,1 */
    U[2][2] = 1.0;
  } else {
    temp = (float)sqrt(temp);
    for (i=0;i<3;i++) {  /* make a unit vector */
      u1[i] /= temp;
      U[i][2] = u1[i];    /* put column in U */
    }
  }
  for (i=0;i<3;i++) {  /* convert v2 to Cartesian */
    u2[i] = 0.0;
    for (j=0;j<3;j++)
      u2[i] += b_mat[i][j] * v2[j];
  }
  u3[0] = u2[1]*u1[2]-u1[1]*u2[2];  /* calculate u2 X u1 */
  u3[1] = u2[2]*u1[0]-u2[0]*u1[2];
  u3[2] = u2[0]*u1[1]-u1[0]*u2[1];
  temp = 0.0;
  for (i=0;i<3;i++)
    temp += u3[i]*u3[i];
  if (temp <= 0.0) {
    u3[0] = 1.0;     /* if null input, make u3 = 1,0,0 */
    U[0][0] = 1.0;
  } else {
    temp = (float)sqrt(temp);
    for (i=0;i<3;i++) {  /* make a unit vector */
      u3[i] /= temp;
      U[i][0] = u3[i];    /* add column to U */
    }
  }
  u2[0] = u1[1]*u3[2]-u3[1]*u1[2];  /* calculate u2 = u1 X u3 */
  u2[1] = u1[2]*u3[0]-u1[0]*u3[2];
  u2[2] = u1[0]*u3[1]-u3[0]*u1[1];
  for (i=0;i<3;i++)
    U[i][1] = u2[i];
/* calculate determinant of U */
  temp =  determinant(U);
  if (temp < 0.0) {
    for (i=0;i<3;i++)
      U[i][0] = - U[i][0];
  }

/* The rotation required to get the desired orientation, which is described
    by G_Rot, has been set up so G_Rot * U = I. Thus G_Rot = U(inv),
    U is unitary and U(inv) = U(transpose). */

  for (i=0;i<3;i++)
    for (j=0;j<3;j++)
      G_Rot[i][j] = U[j][i];
/* calculate rotations */
  temp = G_Rot[1][2] * G_Rot[1][2] + G_Rot[2][2]*G_Rot[2][2];
  if (temp > 0.001) {
    float temp2;
    yrot =  atan(G_Rot[0][2]/sqrt(temp))*RAD;
    temp2 = (float)G_Rot[1][2]/cos(yrot/RAD);
    if (temp2 > 1.0) temp2 = 1.0;
    if (temp2 < -1.0) temp2 = -1.0;
    xrot = -(float)asin(temp2)*RAD;
    temp2 = -(float)G_Rot[0][1]/cos(yrot/RAD);
    if (temp2 > 1.0) temp2 = 1.0;
    if (temp2 < -1.0) temp2 = -1.0;
    zrot = asin(temp2)*RAD;
  } else {
    yrot = -asin(G_Rot[0][2])*RAD;
    xrot = 0.0;      /* XXXXXXXXXXX  - May not be right */
    zrot = 0.0;      /* XXXXXXXXXXX  - May not be right */
  }
}

/* ************************************************************** */
/* ************************************************************** */

void check_atom_label()

/* Routine to verify that each set of anisotropic thermal parameters has
   a corresponding set of atom coordinates.  The atom pointer is set to
   point at those coordinates.  */

{
  int i,j,ind;

  for(i=1;i<n_ellips;i++){
    ind = 0;
    for(j=0;j<natom;++j) {
      if ((atom_l[j][0] == ellips_l[i][0])
       && (atom_l[j][1] == ellips_l[i][1])
       && (atom_l[j][2] == ellips_l[i][2])
       && (atom_l[j][3] == ellips_l[i][3])
       && (atom_n[j] == ellips_n[i])) ind = j+1;
    }
    if (ind == 0) {
      fprintf(flout,"Atom name and number for anisotropic coefficients does not match an atom in the input list.\n");
      fprintf(flout,"Ellipsoid label and number: %c%c%c%c %d.\n"
       ,ellips_l[i][0],ellips_l[i][1],ellips_l[i][2],ellips_l[i][3]
       ,ellips_n[i]);
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    ellips_n[i] = ind-1;
  }
}  /* end of check_atom_label */

/* ************************************************************** */
/* ************************************************************** */

void convert_ellipsoid()

/* routine to convert ellipsoid coefficients Bij or Uij to betaij form */

{
  int i,j;

  for (i=1; i< n_ellips; ++i) {
    if (ell_type[i] != 0){ /* multiply Uij and Bij by a*(i)*a*(j)/4 */
      for (j=0;j<3;++j)ellips[i][j] *= rec_lat_con[j] * rec_lat_con[j] * 0.25;
      ellips[i][3] *= rec_lat_con[0] * rec_lat_con[1] * 0.25;
      ellips[i][4] *= rec_lat_con[0] * rec_lat_con[2] * 0.25;
      ellips[i][5] *= rec_lat_con[1] * rec_lat_con[2] * 0.25;
      if (ell_type[i] > 0) /* multiply Uij by 8pi^2 */
        for (j=0;j<6;++j)ellips[i][j] *= 78.9568;
    }
  }
}  /* end of convert_ellipsoid */

/* ************************************************************** */
/* ************************************************************** */

float dist(j,k)

/* calculates distance between vertices at s_vert[j] and s_vert[k] */

int j,k;

{
    float d,t;
    int i;

    d = 0.0;
    for (i=0;i<=2;++i) {
	t = s_vert[3*j+i] - s_vert[3*k+i];
	d += t*t;
    }
    return((float)sqrt(d));
}

/* ************************************************************** */
/* ************************************************************** */

void do_drawing(output)

GEOMedit_list *output;

/* routine to put together the elements of the drawing */

{

  draw_cell(output,docell);   /* generate unit-cell corners and draw it */
  nvert = 0;                  /* empty vertex list */
  generate_spheres(output);   /* generate all spheres to be drawn within box */
  if (n_ellips > 1 ) {
    generate_ellipsoids(output);  /* now the ellipsoids */
  }
  generate_bonds(output);     /* generate all bonds to be drawn within box */
  generate_poly(output);      /* generate all polyhedra to be drawn */
  generate_planes(output);    /* now the planes */
}


/* ************************************************************** */
/* ************************************************************** */

void expand_atom(natom)

/* routine to expand an atom position and obtain all equipoints within
   the unit cell */

int natom;      /* number of atom in list to expand */

{
  float xp[3],xpp[3];         /* user in generating position */
  int i,j,k,l;                /* loop variables */

  ncell = 0;                  /* clear list */
  for (i=0;i<ng;++i) {        /* symmetry operations */
    for (j=0;j<=2;++j) {
      xp[j] = ts[i][j];
      for (k=0;k<=2;++k)
	xp[j] = xp[j] + ss[i][j][k] * atom_xyz[natom][k];
    }
    for (k=0;k<nlat;++k) {  /*lattice points */
      for (l=0;l<=2;++l)
	xpp[l] = xp[l] + lat_pos[k][l];
      add_to_list(xpp,i);
      if (ns== 0) {
	for (l=0;l<=2;++l)
          xpp[l] = -xpp[l];
	add_to_list(xpp,i);
      }
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

int eigen(biso, beta, valu, vect)

/* routine to calculate eigenvalues 'valu' and vectors 'v' for
   array of anisotropic thermal parameters 'beta', 'biso' is the equivalent
   isotropic B  */

/* routine copied from ORTEP */

float beta[6], valu[3], vect[3][3];
float *biso;

{
  static double errnd = 0.000007;
  double b0,b1,b2,c0,c1;
  double a[3][3],b[3][3],w[3][3],u[3],v[3],z[3];
  double x,y,tem,sigma,vnew,vold,smax;
  int i,j,l,ii,iii,i1,imax;
  static double tpi2 = 19.7392088;  /* 1/2 pi^2 */

#define phif(z) ((b2 - z) * z + b1) * z + b0

/* put beta into square matrix b (called M in ORTEP) */

  for (i=0;i<3;i++) b[i][i] = beta[i];
  b[0][1] = b[1][0] = beta[3];
  b[0][2] = b[2][0] = beta[4];
  b[1][2] = b[2][1] = beta[5];
/* multiply b * ginv to get w */
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      w[i][j] = 0.0;
      for (l=0;l<3;l++) {
        w[i][j] += b[l][i] * ginv[j][l];
      }
    }
  }
  sigma = 0.0;
  for (i=0;i<3;i++) {
    for (j=0;j<3;j++) {
      a[i][j] = w[i][j];
      sigma += a[i][j] * a[i][j];
    }
  }
  if (sigma <= 0.0) return(0);  /* error on null 'w' */
  sigma = sqrt(sigma);
/* get coefficients of third-order characteristic equation 
        Equation is -L^3 + b2 L^2 + b1 L + b0 = 0
*/
  b2 =   a[0][0] + a[1][1] + a[2][2];
  b1 = - a[0][0] * a[1][1] - a[0][0] * a[2][2] - a[1][1] * a[2][2]
       + a[0][2] * a[2][0] + a[0][1] * a[1][0] + a[1][2] * a[2][1];
  b0  =  a[0][0] * a[1][1] * a[2][2] + a[1][0] * a[2][1] * a[0][2]
       + a[2][0] * a[1][2] * a[0][1] - a[2][0] * a[0][2] * a[1][1]
       - a[0][0] * a[2][1] * a[1][2] - a[1][0] * a[0][1] * a[2][2];

  *biso = 4.0*b2/3.0;

/* first root by bisection */

  x = 0.0;
  y = sigma;
  tem = phif(sigma);
  vnew = 0.0;
  if (b0 != 0.0) {
    if ((b0 < 0.0) && (tem <= 0.0)) y = -y;
    if ( b0 > 0.0) {
      y = 0.0;
      x = sigma;
      if (tem > 0.0) x = -x;
    }

/* phif(y) > 0 and phif(x) < 0  */

    for (j=0;j<80;++j) {
      if ((tem = phif(vnew)) == 0.0) break;
      if (tem < 0.0) {
        x = vnew;
      } else y = vnew;
      vold = vnew;
      vnew = (x + y) * 0.5;
      tem = fabs(vold - vnew);
      if (tem <= errnd) break;
      if (vold != 0.0){
        if (fabs(tem/vold) <= errnd ) break;
      }
    }
    if (j == 80){
      fprintf(flout,"No convergence.\n");
      return(0);
    }
  }
  u[2] = vnew;
  c1 = b2 - vnew;
  c0 = b1 + c1 * vnew;
  tem = c1 * c1 + 4.0 * c0;
  if (tem < -0.0001){
    fprintf(flout,"Complex Roots! tem = %f\n",tem);
    return(0);   /* error if complex roots */
  }
  if (tem < 0.0) tem = 0.0;
  tem = sqrt(tem);
  u[0] = 0.5 * (c1 - tem);             
  u[1] = 0.5 * (c1 + tem);
  for (j=0;j<2;j++) { /* sort roots in increasing order */
    if (u[j] > u[2]) {
      tem = u[2];
      u[2] = u[j];
      u[j] = tem;
    }
  }
  if (u[0] < 0.0) {
    fprintf(flout,"Imaginary Root, squares are: %8.5f %8.5f %8.5f\n",u[0],u[1],u[2]);
    return(0);
  }

/* count multiple roots */
  tem = 100.0 * errnd;
  j = 0;
  if ((u[1] - u[0] - tem) < 0) j++;
  if ((u[2] - u[1] - tem) < 0 )j += 2;
  if ( j == 3) {
/* 3 equal roots */
    for (ii=0;ii<3;ii++) {
      for (i=0;i<3;i++)
        vect[i][ii] = 0.0;
      vect[ii][ii] = 1.0;
      valu[ii] = sqrt(u[ii]/tpi2);
    }
    return(1);
  }
  for (ii=0;ii<3;ii++) {
    for (iii=0;iii<3;iii++){
      for (i=0;i<3;i++)
        a[i][iii] = w[i][iii];
    a[iii][iii] = w[iii][iii] - u[ii];
    }
    smax = 0.0;
    for (i=0;i<3;i++) {
      i1 = (i < 2) ? i+1 : 0;
      b[0][i] = a[1][i] * a[2][i1] - a[2][i] * a[1][i1];
      b[1][i] = a[2][i] * a[0][i1] - a[0][i] * a[2][i1];
      b[2][i] = a[0][i] * a[1][i1] - a[1][i] * a[0][i1];
      tem = b[0][i] * b[0][i] + b[1][i] * b[1][i] + b[2][i] * b[2][i];
      if (tem > smax) {
        smax = tem;
        imax = i;
      }
    }
    if (smax <= 0.0) return(0);
    smax = sqrt(smax);
    for (i=0;i<3;i++) v[i] = b[i][imax]/smax;
    axeqb(a,z,v);
/* convert direct-space vector z into Cartesian system */
    for (i=0;i<3;i++) {
      v[i] = 0.0;
      for (j=0;j<3;j++)
        v[i] += b_mat[i][j] * z[j];
    }
    tem = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
    tem = sqrt(tem);
    for (i=0;i<3;i++) vect[ii][i] = v[i]/tem; /* unit vector */
  } 
  for (i=0;i<3;i++) valu[i] = sqrt(u[i]/tpi2);
/* look for degenerate vectors */
  if (fabs(vect[0][0] * vect[1][0] + vect[0][1] * vect[1][1]
     + vect[0][2] * vect[1][2]) > 0.001) {  /* set #1 = 2 x 3 */
     vect[0][0] = vect[1][1] * vect[2][2] - vect[2][1] * vect[1][2];
     vect[0][1] = vect[2][0] * vect[1][2] - vect[1][0] * vect[2][2];
     vect[0][2] = vect[1][0] * vect[2][1] - vect[2][0] * vect[1][1];
  }
  if (fabs(vect[2][0] * vect[1][0] + vect[2][1] * vect[1][1]
     + vect[2][2] * vect[1][2]) > 0.001) {  /* set #3 = 1 x 2 */
     vect[2][0] = vect[0][1] * vect[1][2] - vect[1][1] * vect[0][2];
     vect[2][1] = vect[1][0] * vect[0][2] - vect[0][0] * vect[1][2];
     vect[2][2] = vect[0][0] * vect[1][1] - vect[1][0] * vect[0][1];
  }
  return(1);
}

/* ************************************************************** */
/* ************************************************************** */

void find_all_in_box(i)

/* routine to locate all atoms of type 'i' in the display box */

int i;      /* atom number */

{
  int n,j,k,l;            /* loop indices */
  int sym_no;             /* symmetry number */

  for (n=1;n<NvertM;n++) { /* loop through the Master list */
    j = (int) atom_no[n]/100;    /* extract atom number */
    if (j == i) {
      sym_no = atom_no[n] - 100*j; /* symmetry operator number */
      for (k=0;k<3;k++) {
        o_vert[3*nvert+k] = xypos[3*n+k];
        s_vert[3*nvert+k] = 0.0;
        for (l=0;l<3;l++) {/* calculate cartesian coordinates */
          s_vert[3*nvert+k] += b_mat[k][l] * (xypos[3*n+l] - origin[l]);
        }
      }
      vert_sym_no[nvert++] = sym_no;
      if (nvert > MAX_VERTS) {
        fprintf(stdout,"Too many vertices for dimensions - increase MAX_VERTS.\n");
#ifdef _MAC_
        printf("Use Apple Q to close this window.\n");
#endif
        exit(0);
      }
    }
  }
}   /* end of find_all_in_box */

/* ************************************************************** */
/* ************************************************************** */

void generate_bonds(output)

GEOMedit_list *output;

/* routine to generate bond descriptions and add then to edit list */

{
  int start,same_type;
  int i,j,k;           /* loop counters */
  int N_Bond;          /* Counter for number of bonds of a type */
  int l1;              /* loop counter for matrix operations */
  int nvert1;          /* vertex counter */
  int Bond_Count;      /* Counter for bonds output */
  char bnd1[4],bnd2[4];/* temporary storage of bond labels */
  float d;            /* bond distance */
  float df[3];        /* vector difference */
  float beta;         /* rotation angle */
  float gamma;        /* another rotation angle */
  if (nbond == 1) return; /* exit if no work to be done */

  Bond_Count = 0;
  for (i=1;i<nbond;++i)   {        /* loop through bond types*/
    nvert = 0;                     /* empty the vertex list */
    for (j=0;j<natom;++j) {        /* loop through atom types */
      if (atom_n[j] < 0)  {       /* negative number implies sphere or ellipsoid */
        int o;
        o = -atom_n[j];
        if (o > 10000) {
          o -= 10000;
          if ((bond_l1[i][0] == ellips_l[o][0]) &&
            (bond_l1[i][1] == ellips_l[o][1]) &&
            (bond_l1[i][2] == ellips_l[o][2]) &&
            (bond_l1[i][3] == ellips_l[o][3]))
               find_all_in_box(j);    /* get all atoms of type j */
        } else {
           if ((bond_l1[i][0] == sphere_l[o][0]) &&
               (bond_l1[i][1] == sphere_l[o][1]) &&
               (bond_l1[i][2] == sphere_l[o][2]) &&
               (bond_l1[i][3] == sphere_l[o][3]))
                 find_all_in_box(j);    /* get all atoms of type j */
	}  /* o > 10000 */
      }    /* atom_n[j] < 0 */
    }      /* for (j=0;j<natom;++j) */

    bnd1[0] = bond_l1[i][0];
    bnd1[1] = bond_l1[i][1];
    bnd1[2] = bond_l1[i][2];
    bnd1[3] = bond_l1[i][3];
    nvert1 = nvert;
    if ((bond_l1[i][0] == bond_l2[i][0]) &&
        (bond_l1[i][1] == bond_l2[i][1]) &&
        (bond_l1[i][2] == bond_l2[i][2]) &&
        (bond_l1[i][3] == bond_l2[i][3])) {
      bnd2[0] = bond_l1[i][0];    /* atoms are same type */
      bnd2[1] = bond_l1[i][1];
      bnd2[2] = bond_l1[i][2]; 
      bnd2[3] = bond_l1[i][3];
      same_type = 1;
    }  else {
      bnd2[0] = bond_l2[i][0];
      bnd2[1] = bond_l2[i][1];
      bnd2[2] = bond_l2[i][2];
      bnd2[3] = bond_l2[i][3];
      same_type = 0;
/* atoms i and j are not the same type */
      for (j=0;j<natom;++j) {   /* loop through atom types */
        if (atom_n[j] < 0) {  /* negative number implies sphere */
          int o;
          o = -atom_n[j];
          if (o > 10000) {
            o -= 10000;
            if ((bond_l2[i][0] == ellips_l[o][0]) &&
              (bond_l2[i][1] == ellips_l[o][1]) &&
              (bond_l2[i][2] == ellips_l[o][2]) &&
              (bond_l2[i][3] == ellips_l[o][3]))
                find_all_in_box(j);    /* get all atoms of type j */
          } else {
            if ((bond_l2[i][0] == sphere_l[o][0]) &&
                (bond_l2[i][1] == sphere_l[o][1]) &&
                (bond_l2[i][2] == sphere_l[o][2]) &&
                (bond_l2[i][3] == sphere_l[o][3]))
                  find_all_in_box(j);    /* get all atoms of type j */
          }    /* o > 10000 */
        }      /* atom_n[j] < 0 */
      }        /* for (j=0;j<natom;++j) */
    }          /* if atoms are same type */

    if ((nvert1 > 0) && (nvert > nvert1) || (same_type==1))     {
      N_Bond = 0;
      for (j=0;j<nvert1;++j)  {      /* loop through atoms of type 1 */
        start=nvert1;
        if (same_type==1)start=j;
        for (k=start;k<nvert;++k) {    /* loop on type 2 atoms  */
          d = dist(j,k);          /* get atomic distance */
          if ((d >= bond_min[i]) && (d <= bond_max[i])) {   /* here if distance within range */
            df[0] = df[1] = Bond_Mult * bond_size[i];  /* scaling */
            df[2] = 0.5 * d;
            Bond_Count++;                 /* update count */

            fprintf(fpoutp," /* Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f*/\n",
                bnd1[0],bnd1[1],o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2],
                bnd2[0],bnd2[1],o_vert[3*k],o_vert[3*k+1],o_vert[3*k+2]);
            fprintf(fpoutp," cylinder{<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>, %8.5f \n",
                s_vert[3*j],s_vert[3*j+1],s_vert[3*j+2],s_vert[3*k],
                s_vert[3*k+1],s_vert[3*k+2],df[0]);
            fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_bond_p[i]); 
            N_Bond++;
            if (no_comment == 0) fprintf(fpoutv,"# Bond: %c%c %8.5f %8.5f %8.5f TO %c%c %8.5f %8.5f %8.5f\n",
                bnd1[0],bnd1[1],o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2],
                bnd2[0],bnd2[1],o_vert[3*k],o_vert[3*k+1],o_vert[3*k+2]);
            for (l1=0;l1<3;++l1)
              df[l1] = s_vert[3*k+l1] - s_vert[3*j+l1];
/*
   calculate rotation matrices needed to rotate bond about Z onto Y-Z plane and then about X so that
   bond is parallel to Z.  Finally, apply inverse transformation to cylinder, which started parallel
   to X.  Resulting cylinder will be parallel to vector between atoms.
*/
            beta = atan2(df[0],df[1]+0.0000001);   /* rotation angle about Z (in radians) */
	    gamma = sqrt(df[0]*df[0] + df[1]*df[1]);
	    gamma = atan2(gamma,df[2]);  /* Rotation angle about X */
            fprintf(fpoutv," Separator{\n");
            fprintf(fpoutv,"  Transform{\n ");
            fprintf(fpoutv,"   translation %5.3f %5.3f %5.3f\n  }\n",(s_vert[3*j]+s_vert[3*k])/2., 
               (s_vert[3*j+1]+s_vert[3*k+1])/2.,(s_vert[3*j+2]+s_vert[3*k+2])/2.);               
            fprintf(fpoutv,"  Rotation {\n");
            fprintf(fpoutv,"   rotation 0 0 1 %f\n  }\n",-beta);
            fprintf(fpoutv,"  Rotation {\n");
            fprintf(fpoutv,"   rotation 1 0 0 %f\n  }\n",1.57-gamma);
            fprintf(fpoutv,"  Material {\n   diffuseColor %s\n",col_bond_v[i]);
            fprintf(fpoutv,"  }\n");
            fprintf(fpoutv,"  Cylinder {\n");
            fprintf(fpoutv,"   parts SIDES\n");
            fprintf(fpoutv,"   radius %f\n",Bond_Mult * bond_size[i]);
            fprintf(fpoutv,"   height %f\n",d);
            fprintf(fpoutv,"  }\n }\n");
          }    /* if distance OK */
        }      /* for k */
      }        /* for j */
    }          /* if nvert */
  }            /* for i */
  printf("%4d bonds.\n",Bond_Count);
  fprintf(flout,"\nGenerated %4d bonds.\n\n",Bond_Count);
}              /* end of generate_bonds */

/* ************************************************************** */
/* ************************************************************** */

void get_atom_id()

/* routine to rework atom numbers to point at polyhedron, sphere
   plane, or ellipsoid characteristics list */

{
  int i,j,ind;                /* local variables */


  for (i=0;i<natom;++i) {
    sv_atom_n[i]=atom_n[i];    /* save original atom number */ 
    ind = 0;
    for (j=1;j<nsphere;++j) {
      if((atom_l[i][0] == sphere_l[j][0])
      && (atom_l[i][1] == sphere_l[j][1])   
      && (atom_l[i][2] == sphere_l[j][2])
      && (atom_l[i][3] == sphere_l[j][3])) {
        atom_n[i] = -j; /* atom is in sphere lst - no. is negative */
        ind = 1;        /* set indicator */
      }
    }
    if (ind == 0) {
      for (j=1;j<npoly;++j) {
        if((atom_l[i][0] == poly_l[j][0])
        && (atom_l[i][1] == poly_l[j][1])   
        && (atom_l[i][2] == poly_l[j][2])
        && (atom_l[i][3] == poly_l[j][3])) {
          atom_n[i] = j;      /* atom is in poly. list */
          ind = 1;            /* set indicator */
        }
      }
    }
    if (ind == 0) {
      for (j=1;j<nplane;++j) {
        if((atom_l[i][0] == plane_l[j][0])
        && (atom_l[i][1] == plane_l[j][1])   
        && (atom_l[i][2] == plane_l[j][2])
        && (atom_l[i][3] == plane_l[j][3])) {
          atom_n[i] = j+10000;      /* atom is in plane list */
          ind = 1;            /* set indicator */
        }
      }
    }
    if(ind == 0) atom_n[i] = 0;  /* if not sphere or polyhedron - use
					default polyhedron (#0) */
  }

  for (j=1;j<n_ellips;++j) {   /* Ellipsoids are special kind of sphere!!! */

      if(atom_n[ellips_n[j]] <=0 )    atom_n[ellips_n[j]] = -10000 - j; /* but polyhedra take precedence over ellipsoids*/
  }    
}     /* end of get_atom_id  */

/* ************************************************************** */
/* ************************************************************** */
 
void get_input()

/* Routine to process input file, make lattice metric and do preliminary
   ellipsoid processing */

{
  int i,j,k,npd;
  float biso;

  fprintf(flout,"\n********************* Input: ************************\n");
  read_inp();
  fprintf(flout,"*****************************************************\n\n");
  print_sym();          /* print symmetry information */
  fclose(fpin);         /* close file */             
  make_bmat(lat_con);   /* create the lattice metric */
/* set limits of search regions */
  if (packflag == 0 && boxflag == 0) { /* neither given */
    cryst_lim[0] = origin[0] - 0.5;
    cryst_lim[1] = origin[1] - 0.5;
    cryst_lim[2] = origin[2] - 0.5; /* set to generate one cell */
    cryst_lim[3] = origin[0] + 0.5;
    cryst_lim[4] = origin[1] + 0.5;
    cryst_lim[5] = origin[2] + 0.5;
    packflag = 1;
  }
  if (packflag != 0) {  /* pack command given - calculate approximate bounds */
    float temp,temp1;
    int i;                                                                  
    for (i=0;i<3;i++) {
      temp = origin[i] - cryst_lim[i];
      temp1 = cryst_lim[i+3] - origin[i];
      if (temp1 > temp) temp = temp1;
      boxlim[i] = (temp + 0.2) * lat_con[i]/0.5;
    }
  } else {    /* bounds command given - calculate approximate pack limits */
    int i;
    for (i=0;i<3;i++) {
      cryst_lim[i] = origin[i] - boxlim[i] / lat_con[i] - 0.2;
      cryst_lim[i+3] = boxlim[i] / lat_con[i] - origin[i] + 0.2;
    }
  }
  if ((Options & L_OPT)!=0)Calc_Rot(lookat_v1,lookat_v2);/* Get rotation angles for "lookat" */
  if (n_ellips > 1) {
    convert_ellipsoid();/* change Bij and Uij forms to betaij */
    npd = 0;
    fprintf(flout,"\nAnalysis of Thermal Ellipsoids:\n");
    for (i=1;i < n_ellips;i++){
      if (eigen(&biso,ellips[i],ellips_RMS[i],ellips_EV[i]) == 0) {
        fprintf(flout,"Ellipsoid for %c%c%c%c %d is non-positive definite.\n"
            ,ellips_l[i][0],ellips_l[i][1],ellips_l[i][2],ellips_l[i][3]
            ,ellips_n[i]);
        npd = 1;
        ellips_RMS[i][0] = sqrt(biso/78.957);      /* convert Biso to mu */
      } else {
        fprintf(flout,"\n%c%c%c%c %3d Equivalent Isotropic B: %6.2f, U: %6.3f\n"
          ,ellips_l[i][0],ellips_l[i][1],ellips_l[i][2],ellips_l[i][3]
          ,ellips_n[i],biso,biso/78.957);
        fprintf(flout,"      RMS Amplitudes: %6.3f %6.3f %6.3f\n"
          ,ellips_RMS[i][0],ellips_RMS[i][1],ellips_RMS[i][2]);
        fprintf(flout,"      Eigen Vectors (by row):\n");
        for (j=0;j<3;j++) {
          fprintf(flout,"     ");
          for (k=0;k<3;k++)
	    fprintf(flout,"     %8.5f",ellips_EV[i][j][k]);
          fprintf(flout,"\n");
        }
      }
    }
    if (npd == 1) {
      n_ellips = 1;
      fprintf(flout,"\n***** Ellipsoid generation is turned off.\n");
    } else {
      P_to_C(Ellipsoid_Prob);
    }
    check_atom_label(); /* Verify that there are xyz's for each bij set */
  }
}   /* end of get_input */

/* ************************************************************** */
/* ************************************************************** */
 
void get_label(input,c1,c2,c3,c4,strip) 

/* Routine to get 1 to 4 character atom ID from input line, and shift input
   line to left  */

char input[100],*c1,*c2,*c3,*c4;
int strip;

{
 
  int i,j;
  i = 0;
  if (strip != 0) 
    while (input[i] != ' ') input[i++] = ' ';   /* skip over command */        
  while (input[i] == ' ') i++;        /* skip whitespace after command */      
  *c1 = input[i];    /* return first atom label character */
  input[i++] = ' ';
  *c2 = input[i];   /* return 2nd atom label char */
  if (input[i] != ' ') input[i++] = ' ';  /* if 2nd not blank, rub it out */
  *c3 = input[i];   /* return 3rd character */
  if (input[i] != ' ') input[i++] = ' ';  /* if 3rd not blank, rub it out */
  *c4 = input[i];   /* return 4th character */
  if (input[i] != ' ') input[i++] = ' ';  /* if 4th not blank, rub it out */
  while (input[i] == ' ') i++;  /* skip white space afer atom label */
  j = 0;   
  for(i;i<100;i++){
    input[j++] = input[i];  /* copy rest of command */
  }
}  /* end of get_label */ 

/* ************************************************************** */
/* ************************************************************** */

void Locate_Triple()

/* routine to try to place the vector triple where it will be seen in the
POV diagram */

{
  int i,j;
  float temp,pos[3],cpos[3];

  for (i=0;i<3;++i) {  /* convert origin to Cartesian */
    pos[i] = 0.0;
    for (j=0;j<3;++j)                             
      pos[i] -= b_mat[i][j] * origin[j];
  }
  for(i=0;i<=2;++i){       /* calculate position of point after POV rotation */
    cpos[i] = 0.0;
    for(j=0;j<=2;++j)
      cpos[i] += G_Rot[j][i] * pos[j];
  }
  for (i=0;i<3;i++) {
    temp = 0.08 * (POV_Max[i] - POV_Min[i]);
    if (cpos[i] > 0)
      pos[i] = POV_Max[i] + temp;
    else
      pos[i] = POV_Min[i] - temp;
  }
/* calculate coordinates of pos before rotation */
  for (i=0;i<3;i++) {
    offset[i] = 0.0;
    for (j=0;j<3;j++) {
      offset[i] += G_Rot[i][j] * pos[j];
    }
  }
}

/* ************************************************************** */
/* ************************************************************** */

void make_bmat(lat_con)

float lat_con[6];

{
    float rad;
    float snal,snbe,snga,csal,csbe,csga,vol,det;
    int i,j;
 
    rad = 45.0/atan(1.0);
    for (i=3;i<=5;++i) if(lat_con[i] <= 0) lat_con[i] = 90.0;
    if (sys == 5) lat_con[5] = 120.0;
    for (i=0;i<=2;++i)
	for (j=0;j<=2;++j) b_mat[j][i] = 0.0; /* initialize matrix */
    b_mat[0][0] = 1.0;
    csal = cos(lat_con[3]/rad);
    snal = sin(lat_con[3]/rad);
    csbe = cos(lat_con[4]/rad);
    snbe = sin(lat_con[4]/rad);
    csga = cos(lat_con[5]/rad);
    snga = sin(lat_con[5]/rad);
    for (i=0;i<3;i++) ginv[i][i] = lat_con[i] * lat_con[i];
    ginv[0][1] = ginv[1][0] = lat_con[0] * lat_con[1] * csga;
    ginv[0][2] = ginv[2][0] = lat_con[0] * lat_con[2] * csbe;
    ginv[1][2] = ginv[2][1] = lat_con[1] * lat_con[2] * csal;
    b_mat[0][1] = csga;
    b_mat[1][1] = snga;
    b_mat[0][2] = csbe;
    b_mat[1][2] = (csal-csbe*csga)/snga;
    b_mat[2][2] = sqrt(1.0 - b_mat[0][2] * b_mat[0][2] - b_mat[1][2] * b_mat[1][2]);
    for (i=0;i<=2;++i)
	for (j=0;j<=2;++j) b_mat[j][i] *= lat_con[i];
/* calculate cell volume and reciprocal lengths */
    vol = b_mat[0][0] * (b_mat[1][1] * b_mat[2][2] - b_mat[1][2] * b_mat[2][1])
        -b_mat[0][1] * (b_mat[1][0] * b_mat[2][2] - b_mat[1][2] * b_mat[2][0])
        +b_mat[0][2] * (b_mat[1][0] * b_mat[2][1] - b_mat[1][1] * b_mat[2][0]);
    rec_lat_con[0] = lat_con[1] * lat_con[2] * snal / vol;
    rec_lat_con[1] = lat_con[0] * lat_con[2] * snbe / vol;
    rec_lat_con[2] = lat_con[1] * lat_con[0] * snga / vol;

/* get binv */

    det =  determinant(b_mat);
    b_inv_mat[0][0] =  (b_mat[1][1] * b_mat[2][2] - b_mat[1][2] * b_mat[2][1])/det;
    b_inv_mat[1][0] = -(b_mat[1][0] * b_mat[2][2] - b_mat[1][2] * b_mat[2][0])/det;
    b_inv_mat[2][0] =  (b_mat[1][0] * b_mat[2][1] - b_mat[1][1] * b_mat[2][0])/det;
    b_inv_mat[0][1] = -(b_mat[0][1] * b_mat[2][2] - b_mat[0][2] * b_mat[2][1])/det;
    b_inv_mat[1][1] =  (b_mat[0][0] * b_mat[2][2] - b_mat[0][2] * b_mat[2][0])/det;
    b_inv_mat[2][1] = -(b_mat[0][0] * b_mat[2][1] - b_mat[0][1] * b_mat[2][0])/det;
    b_inv_mat[0][2] =  (b_mat[0][1] * b_mat[1][2] - b_mat[0][2] * b_mat[1][1])/det;
    b_inv_mat[1][2] = -(b_mat[0][0] * b_mat[1][2] - b_mat[0][2] * b_mat[1][0])/det;
    b_inv_mat[2][2] =  (b_mat[0][0] * b_mat[1][1] - b_mat[0][1] * b_mat[1][0])/det;
}  /* end of make_bmat */
     
/* ************************************************************** */
/* ************************************************************** */

void Output_Spheres(radii,output,i)

/* routine to output the spheres with vertices defined in s_vert and o_vert
   The radius of the sphere is in radii */

float *radii;
GEOMedit_list *output;
int i;

{
  int k;
  for(k=0;k<nvert;++k){
    fprintf(fpoutp," /* Sphere for %c%c%c%c at %8.5f %8.5f %8.5f */ \n"
      ,sphere_l[i][0],sphere_l[i][1],sphere_l[i][2],sphere_l[i][3]
      ,o_vert[3*k],o_vert[3*k+1],o_vert[3*k+2]);
    fprintf(fpoutp," sphere{<%8.5f, %8.5f, %8.5f>, %8.5f\n",s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2],radii[k]);
    fprintf(fpoutp,"  texture{pigment{color %s  }}\n",sphere_col_p[i]);
    fprintf(fpoutp,"   finish{phong %5.2f phong_size %5.2f reflection 0.1}\n"
      ,Phong_Value,Phong_Size);
    fprintf(fpoutp," }\n");
    if (no_comment == 0)fprintf(fpoutv,"# Sphere for %c%c%c%c at %8.5f %8.5f %8.5f \n"
      ,sphere_l[i][0],sphere_l[i][1],sphere_l[i][2],sphere_l[i][3]
      ,o_vert[3*k],o_vert[3*k+1],o_vert[3*k+2]);
    fprintf(fpoutv," Separator {");
    fprintf(fpoutv,"  Transform {");
    fprintf(fpoutv," translation %8.5f %8.5f %8.5f}\n",s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
    fprintf(fpoutv,"  Material {");
    fprintf(fpoutv," shininess 0.3\n");
    fprintf(fpoutv," diffuseColor %s}\n",sphere_col_v[i]);
    fprintf(fpoutv,"  Sphere { radius %8.5f}\n",radii[k]);
    fprintf(fpoutv," }\n");
  }
}

/* ************************************************************** */
/* ************************************************************** */

float P_to_C(prob)

/* Routine to convert probability to ellipsoid scaling, C.  Given probability
   'prob', calculate value of C required for ellipsoid,

   C^2 = (delx/sigx)^2 + (dely/sigy)^2 + (delz/sigz)^2,

   where 'prob' is the probability that a random point in the distribution
   will fall inside the ellipsoid.  Table values taken from D.B. Owen,
   "Handbook of Statistical Tables", Addison-Wesley, Reading, Mass., 1962.

   N.B. The present version of the tables only handles probabilities from
   0.01 to 0.99.
*/

float prob;

#define N_C 99

{
  static float C_table[N_C+1] =
    {0.3389,0.4299,0.4951,0.5479,0.5932,0.6334,0.6699,0.7035,0.7349,0.7644,
     0.7924,0.8192,0.8447,0.8694,0.8932,0.9162,0.9386,0.9605,0.9818,1.0026,
     1.0230,1.0430,1.0627,1.0821,1.1012,1.1200,1.1386,1.1570,1.1751,1.1932,
     1.2110,1.2288,1.2464,1.2638,1.2812,1.2985,1.3158,1.3330,1.3501,1.3672,
     1.3842,1.4013,1.4183,1.4354,1.4524,1.4695,1.4866,1.5037,1.5209,1.5382,
     1.5555,1.5729,1.5904,1.6080,1.6257,1.6436,1.6616,1.6797,1.6980,1.7164,
     1.7351,1.7540,1.7730,1.7924,1.8119,1.8318,1.8519,1.8724,1.8932,1.9144,
     1.9360,1.9580,1.9804,2.0034,2.0269,2.0510,2.0757,2.1012,2.1274,2.1544,
     2.1824,2.2114,2.2416,2.2730,2.3059,2.3404,2.3767,2.4153,2.4563,2.5003,
     2.5478,2.5997,2.6571,2.7216,2.7955,2.8829,2.9912,3.1365,3.3682,5.9503};

  int i;
  float C;
  i = 100.0 * prob;    /* find correct table entry */
  if ( i > N_C) i = N_C;
/* interpolate (linear) */
  C = C_table[i-1] + (prob - 0.01 * i) * (C_table[i] - C_table[i-1]); 
  return (C);
}

/* ************************************************************** */
/* ************************************************************** */

void plot_vrml_poly()

{
  int i,j,k,l,m,n ;                         
  float *out;
  int *point,equal;

/* get space for unique vertices and a list to point to them */

  if ((out = (float *) malloc((unsigned) (draw_list * 3 * sizeof(float)))) == NULL){
    fprintf(stderr,"Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  if ((point = (int *) malloc((unsigned) (draw_list * sizeof(int)))) == NULL){
    free(out);
    fprintf(stderr,"Unable to allocate space in plot_vrml_poly.\n");
    return;
  }
  fprintf(fpoutv,"Coordinate3 { point [\n");
  n=0;
  for (i=0;i<draw_list;i++){   /* build list of unique points */
    j=poly_list[i];
/*    fprintf(fpoutv,"# %d-sided polygon:\n",j); */
    for (k=i+1;k<=i+j;k++){
      l=poly_list[k]-1;
      for(m=0;m<3;m++)
        out[3*n+m] = s_vert[3*l+m];
      equal = 0;
      for (m=0;m<n;m++)       /* check if new */
        if ((fabs(out[3*m]   - out[3*n])   < 0.0002)
          &&(fabs(out[3*m+1] - out[3*n+1]) < 0.0002)
          &&(fabs(out[3*m+2] - out[3*n+2]) < 0.0002)) {
            point[k] = m;    /* point previously found */
            equal = 1;
            break;
        }
      if (equal == 0)
        point[k] = n++;         /* new point - save it by incrementing n */
    }
    i=i+j;
  }
  --n;
  for (l=0;l<n;l++)           /* output the unique set of coordinates */
    fprintf(fpoutv," %5.3f %5.3f %5.3f,\n",out[3*l],out[3*l+1],out[3*l+2]);
  fprintf(fpoutv," %5.3f %5.3f %5.3f]\n",out[3*n],out[3*n+1],out[3*n+2]);
  fprintf(fpoutv,"}\n IndexedFaceSet { coordIndex [\n");
  l=0;
  for (i=0;i<draw_list-1;i++){
    j=poly_list[i];
    l++;
/*    fprintf(fpoutv,"# %d-sided polygon:\n",j);     */
    for (k=0;k<j;k++)
      fprintf(fpoutv,"%d,",point[l++]);
    if (i+j<draw_list-2){        
      fprintf(fpoutv,"-1,\n");
    }else{
      fprintf(fpoutv,"-1]\n  }\n");
    }
    i=i+j;
  }
  if (edges==1){
    fprintf(fpoutv," Material {\n   diffuseColor %s \n }\n",col_edge_v);
    fprintf(fpoutv,"IndexedLineSet { coordIndex [\n");
    l=0;
    for (i=0;i<draw_list-1;i++){
      j=poly_list[i];
      l++;
      for (k=0;k<j;k++)
        fprintf(fpoutv,"%d,",point[l++]);
      if (i+j<draw_list-2){        
        fprintf(fpoutv,"%d,-1,\n",point[l-j]);
      }else{
        fprintf(fpoutv,"%d,-1]\n  }\n }\n",point[l-j]);
      }
      i=i+j;
    }
  }else{
    fprintf(fpoutv," }\n");
  }
  free(out);              /* release the space */
  free(point);
  return;
}

/* ************************************************************** */
/* ************************************************************** */

void print_sym()

/* print symmetry information */

{
  int i,j,k;

  switch (sys)  {
    case 1: fprintf(flout,"Triclinic - ");break;
    case 2: fprintf(flout,"Monoclinic - ");break;
    case 3: fprintf(flout,"Orthorhombic - ");break;
    case 4: fprintf(flout,"Tetragonal - ");break;
    case 5: fprintf(flout,"Hexagonal - ");break;
    case 6: fprintf(flout,"Cubic - ");
  }
  if (ns == 0) {
    fprintf(flout,"Centric - ");
  } else {
    fprintf(flout,"Acentric - ");
  }
  switch (nbr) {
    case 1: fprintf(flout,"P lattice"); break;
    case 2: fprintf(flout,"A-centered");break;
    case 3: fprintf(flout,"B-centered");break;
    case 4: fprintf(flout,"C-centered");break;
    case 5: fprintf(flout,"F-centered");break;
    case 6: fprintf(flout,"I-centered");break;
    case 7: fprintf(flout,"R-centered");break;
    default: fprintf(flout,"**** Unknown Lattice ***");
  }
  fprintf(flout," with %d Symmetry Operators\n\n",ng);
  for (i=1;i<=ng;++i) {
    for (j=1;j<=3;++j) {
      for (k=1;k<=3;++k) {
	if (ss[i-1][j-1][k-1] != 0) {
          if(ss[i-1][j-1][k-1] > 0) { fprintf(flout,"+");
	} else {fprintf(flout,"-");}
        switch (k) {
          case 1: fprintf(flout,"x");break;
	  case 2: fprintf(flout,"y");break;
	  case 3: fprintf(flout,"z");
	}
      }
    }
    if(ts[i-1][j-1] != 0.0) {
      if(ts[i-1][j-1] == 0.5) fprintf(flout,"+1/2");
      if(ts[i-1][j-1] == 0.25) fprintf(flout,"+1/4");
      if(ts[i-1][j-1] == 0.75) fprintf(flout,"+3/4");
      if(fabs(ts[i-1][j-1] - 0.33333333) < 0.00001 ) fprintf(flout,"+1/3");
      if(fabs(ts[i-1][j-1] - 0.66666667) < 0.00001 ) fprintf(flout,"+2/3");
      if(fabs(ts[i-1][j-1] - 0.16666667) < 0.00001 ) fprintf(flout,"+1/6");
      if(fabs(ts[i-1][j-1] - 0.83333333) < 0.00001 ) fprintf(flout,"+5/6");
    }
    if (j != 3) fprintf(flout,",");
  }
  if (i != ng)
    if (i % 4 != 0) {fprintf(flout," ; ");} else {fprintf(flout,"\n");}
 }
 fprintf(flout,"\n");
}    /* end of print_sym */

/* ************************************************************** */
/* ************************************************************** */

void read_inp()

/* routine to read and process input file */

{
  char input[100];               /* input buffer for data read from file */
  char string[10];               /* temporary string */
  int intype;                    /* type of datum read */
  int i,j;                       /* temporary */
  char t_name[4];                /* temporary atom name storage */
  char t_color[30];              /* temporary ellipsoid color storage */
  int in_line;			 /* flag for inline import of foreign files */
  float xyzoff[3];               /* amount to be added to each atom position */

/* ************************************************************** */
/* ************************************************************** */
/*
    Commands Implemented

    atom
    betaij
    bij,Bij
    bond
    bounds
    box
    cell
    cutout
    depthcue
    edges
    ellipcolor
    ellipsoids
    import
    inline
    lookat
    magnification
    molcomp
    origin
    pack
    phong
    plane
    polysz
    polyfudge  (no longer needed)
    rem
    sgrp, spgp, spgr
    sphere
    title, titl
    uij,Uij
    vectors    
    view
    xyzoff
    end, END

                                                                  */
/* ************************************************************** */
/* ************************************************************** */

  for (i=0;i<3;i++) xyzoff[i] = 0.0;
  while (!feof(fpin)) {
    if(fgets(input,100,fpin)) {             /* read a line */
      Blank_Strip(input);          /* suppress any input blanks */
      fprintf(flout,"*         %s",input);  /* echo to listing file */
    } else {
      fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(1);
    }
    intype = 0;
    if(strncmp(input,"title",5) == 0)intype = 1;
    if(strncmp(input,"titl",4) == 0)intype = 1;
    if(strncmp(input,"rem",3) == 0)intype = 1;  /* rem only echos to output */
    if(strncmp(input,"cell",4) == 0)intype = 2;
    if(strncmp(input,"sgrp",4) == 0)intype = 3;
    if(strncmp(input,"spgp",4) == 0)intype = 3;
    if(strncmp(input,"spgr",4) == 0)intype = 3;
    if(strncmp(input,"ellipsoids",10) == 0)intype = 4;
    if(strncmp(input,"lookat",6) == 0)intype = 5;
    if(strncmp(input,"atom",4) == 0)intype = 6;
    if(strncmp(input,"polysz",6) == 0)intype = 7;
    if(strncmp(input,"sphere",6) == 0)intype = 8;
    if(strncmp(input,"end",3) == 0)intype = 9;
    if(strncmp(input,"END",3) == 0)intype = 9;
    if(strncmp(input,"xyzoff",6) == 0)intype = 10;
    if(strncmp(input,"bond",4) == 0)intype = 11;
    if(strncmp(input,"box",3) == 0)intype = 12;
    if(strncmp(input,"edges",5) == 0)intype = 13;
    if(strncmp(input,"view",4) == 0) intype =14;
    if(strncmp(input,"plane",5)==0) intype=15;
    if(strncmp(input,"bij",3)==0) intype=16;
    if(strncmp(input,"Bij",3)==0) intype=16;
    if(strncmp(input,"betaij",6)==0) intype=17;
    if(strncmp(input,"uij",3)==0) intype=18;
    if(strncmp(input,"Uij",3)==0) intype=18;
    if(strncmp(input,"Uij",3)==0) intype=18;
    if(strncmp(input,"bounds",6)==0) intype=19;
    if(strncmp(input,"pack",4)==0) intype=20;
    if(strncmp(input,"magnification",13)==0) intype=21;
    if(strncmp(input,"import",6) == 0)intype = 22;
    if(strncmp(input,"origin",6) == 0)intype = 23;
    if(strncmp(input,"ellipcolor",10) == 0)intype = 24;
    if(strncmp(input,"inline",6) == 0)intype = 25;
    if(strncmp(input,"polyfudge",9) == 0) intype = 26;
    if(strncmp(input,"vectors",7) == 0) intype = 27;
    if(strncmp(input,"phong",5) == 0) intype = 28;
    if(strncmp(input,"molcomp",7) == 0) intype = 29;
    if(strncmp(input,"cutout",6) == 0) intype = 30;
    if(strncmp(input,"depthcue",8) == 0) intype = 31;

    switch (intype) {
    int done;
        case 1: break;      /* ignore title except for listing */
        case 2: for (i=3;i<=5;++i) lat_con[i] = 0.;
            sscanf(input,"%s %f %f %f %f %f %f",string,&lat_con[0]
	      ,&lat_con[1],&lat_con[2],&lat_con[3],&lat_con[4]
 	      ,&lat_con[5]);
	    break;
	case 3: symop(input);   /* process space group information */
	    break;
        case 4: 
            do_ellipsoids = 1;
            sscanf(input,"%s %f",string,&Ellipsoid_Prob);
            if (Ellipsoid_Prob > 1) Ellipsoid_Prob *= 0.01; /* Probability to fraction */
            break;
        case 5: /* read vectors for lookat command */
            sscanf(input,"%s %f %f %f %f %f %f",string,&lookat_v1[0]
              ,&lookat_v1[1],&lookat_v1[2],&lookat_v2[0],&lookat_v2[1]
              ,&lookat_v2[2]);
            Options = Options | L_OPT;
	    break;
        case 6: get_label(input,&atom_l[natom][0],&atom_l[natom][1]
                   ,&atom_l[natom][2],&atom_l[natom][3],1);
            sscanf(input,"%d %f %f %f",
		 &atom_n[natom],&atom_xyz[natom][0],
                 &atom_xyz[natom][1],&atom_xyz[natom][2]);
	    natom++;
            if (natom >= MAX_ATOM) {
              fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
	    break;
        case 7: get_label(input,&poly_l[npoly][0],&poly_l[npoly][1]
                   ,&poly_l[npoly][2],&poly_l[npoly][3],1);
            sscanf(input,"%f %30c",
                 &poly_size[npoly],poly_col_p[npoly]);
            strcpy(poly_col_v[npoly],poly_col_p[npoly]);
            check_color(poly_col_v[npoly]);
	    npoly++;
            if (npoly >= MAX_POLY) {
              fprintf(flout,"Too many polyhedra specified - increase MAX_POLY.\n");
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
	    break;
        case 8: get_label(input,&sphere_l[nsphere][0],&sphere_l[nsphere][1]
                   ,&sphere_l[nsphere][2],&sphere_l[nsphere][3],1);
            sscanf(input,"%f %30c",
                 &sphere_size[nsphere],sphere_col_p[nsphere]);
            strcpy(sphere_col_v[nsphere],sphere_col_p[nsphere]);
            check_color(sphere_col_v[nsphere]);
	    nsphere++;
            if (nsphere >= MAX_SPHERE) {
              fprintf(flout,"Too many sphere types specified - increase MAX_SPHERE.\n");
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
	    break;
        case 9:
            if (n_ellips > 1) do_ellipsoids = 1;
            for (i=0;i<natom;i++)     /* shift atom coordinates for origin offset */
              for (j=0;j<3;j++)
                atom_xyz[i][j] -= xyzoff[j];
            return;     /* thru with end */
            if (packflag != 0) {
              boxlim[0] = boxlim[1] = boxlim[2] = 500.0;
            }
        case 10: sscanf(input,"%s %f %f %f",string,&xyzoff[0]
              ,&xyzoff[1],&xyzoff[2]);
            break;
        case 11: get_label(input,&bond_l1[nbond][0],&bond_l1[nbond][1]
                     ,&bond_l1[nbond][2],&bond_l1[nbond][3],1);
                 get_label(input,&bond_l2[nbond][0],&bond_l2[nbond][1]
                     ,&bond_l2[nbond][2],&bond_l2[nbond][3],0);
                 sscanf(input,"%f %f %f %30c",
		   &bond_size[nbond],&bond_min[nbond],&bond_max[nbond],
                   col_bond_p[nbond]);
            if (strlen(col_bond_p[nbond]) == 0) strcpy(col_bond_p[nbond],"Grey");
            strcpy(col_bond_v[nbond],col_bond_p[nbond]);
            check_color(col_bond_v[nbond]);
	    nbond++;
            if (nbond >= MAX_BOND) {
              fprintf(flout,"Too many bond types specified - increase MAX_BOND.\n");
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
	    break;
        case 12: sscanf(input,"%3s %f %30c",string,&rad_cell,col_cell_p);
            if (strlen(col_cell_p) == 0) strcpy(col_cell_p,"Black");
            strcpy(col_cell_v,col_cell_p);
            check_color(col_cell_v);
            if (rad_cell == 0.0)docell = 0;
	    break;
        case 13: sscanf(input,"%5s %f %30c",string,&rad_edge,col_edge_p);
            edges=1;
            if (rad_edge == 0) rad_edge=0.01;
            if (strlen(col_edge_p) == 0) strcpy(col_edge_p,"Black");
            strcpy(col_edge_v,col_edge_p);
            check_color(col_edge_v);
	    break;	    
        case 14:
            if ((Options & V_OPT)==0)  /* only if no -v switch */
              sscanf(input,"%4s %f %f %f",string,&xrot,&yrot,&zrot);
	    break;	
        case 15: get_label(input,&plane_l[nplane][0],&plane_l[nplane][1]
                   ,&plane_l[nplane][2],&plane_l[nplane][3],1);
                 sscanf(input,"%f %30c",
                   &plane_size[nplane],plane_col_p[nplane]);
            strcpy(plane_col_v[nplane],plane_col_p[nplane]);
            check_color(plane_col_v[nplane]);
	    nplane++;
            if (nplane >= MAX_PLANE) {
              fprintf(flout,"Too many plane types specified - increase MAX_PLANE.\n");
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
	    break;
        case 16: ell_type[n_ellips] = -1;   /* Bij */
        case 17: if (intype == 17) ell_type[n_ellips] = 0; /* betaij */
        case 18: if (intype == 18) ell_type[n_ellips] = 1; /* Uij */
            get_label(input,&ellips_l[n_ellips][0],&ellips_l[n_ellips][1]
               ,&ellips_l[n_ellips][2],&ellips_l[n_ellips][3],1);
            sscanf(input,"%d %f %f %f %f %f %f %30c",&ellips_n[n_ellips],
              &ellips[n_ellips][0],&ellips[n_ellips][1],&ellips[n_ellips][2],
              &ellips[n_ellips][3],&ellips[n_ellips][4],&ellips[n_ellips][5],
              ellips_col_p[n_ellips]);   /* read unique coefficients and color */
            strcpy(ellips_col_v[n_ellips],ellips_col_p[n_ellips]);
            if ( strlen(ellips_col_v[n_ellips]) != 0) check_color(ellips_col_v[n_ellips]);
            n_ellips++;
            if (n_ellips >= MAX_EL) {
              fprintf(flout,"Too many ellipsoid types specified - increase MAX_EL.\n");
#ifdef _MAC_
              printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
            break;
        case 19:    /* bounds command */
            if ((Options & B_OPT) == 0) {  /* only do this if no -b switch */
              sscanf(input,"%s %f %f %f",string,&boxlim[0],&boxlim[1],&boxlim[2]);
              boxflag = 1;
            }
            break;
        case 20:    /* pack command */
            if ((Options & P_OPT)==0) { /* only if no -p switch */
              sscanf(input,"%s %f %f %f %f %f %f",string,&cryst_lim[0],&cryst_lim[3]
                ,&cryst_lim[1],&cryst_lim[4],&cryst_lim[2],&cryst_lim[5]);
              packflag = 1;
            }
            break;
        case 21:
            if ((Options & M_OPT)==0)  /* only if no -m switch */
              sscanf(input,"%s %f",string,&Magnification);
            break;
        case 22:
        case 25:   /* import command */
            in_line =0;
            done = 0;
            if (strncmp(input,"inline",6)==0)in_line=1;
            Token_Strip(input,1);      /* strip command */
            if(strncmp(input,"gsas",4)==0){
              import_gsas(input,in_line);
              done=1;
            }
            if(strncmp(input,"cif",3)==0){
              import_cif(input,in_line);
              done=1;
            }
            if(strncmp(input,"schakal",7)==0){
              import_schakal(input,in_line);
              done=1;
            }
            if(strncmp(input,"shakal",6)==0){
              import_schakal(input,in_line);
              done=1;
            }
            if(strncmp(input,"shelx",5)==0){
              import_shelx(input,in_line);
              done=1;
            }
            if(strncmp(input,"fdat",4)==0){
              import_fdat(input,in_line);
              done=1;
            }
            if(strncmp(input,"csd",3)==0){
              import_fdat(input,in_line);
              done=1;
            }
            if(strncmp(input,"ccdf",4)==0){
              import_fdat(input,in_line);
              done=1;
            }
            if (done == 0) {
              fprintf(flout,"Import/inline source type '%s' not recognized.\n",input);
            }
            break;
        case 23:    /* origin command */
            if ((Options & O_OPT)==0) { /* only if no -o switch */
              sscanf(input,"%s %f %f %f",string,&origin[0],&origin[1],&origin[2]);
            }
            break;
        case 24:    /* ellipcolor command */
            get_label(input,&t_name[0],&t_name[1],&t_name[2],&t_name[3],1);
            if (input[0] == '*') {
              sscanf(input,"%s %s",string,t_color);
              j = -1;
            } else {
              sscanf(input,"%d %s",&j,t_color);
            }
            for(i=1;i<n_ellips;i++) { /* find this atom in ellipsoid list */
              if ((t_name[0] == ellips_l[i][0]) &&
                  (t_name[1] == ellips_l[i][1]) &&
                  (t_name[2] == ellips_l[i][2]) &&
                  (t_name[3] == ellips_l[i][3])) {
                if ((j == -1) || (j == ellips_n[i])) {
                   strcpy(ellips_col_p[i],t_color);             /* copy color */
                   strcpy(ellips_col_v[i],ellips_col_p[i]);
                   check_color(ellips_col_v[i]);
                }
              }
            }
            break;
        case 26: /* polyfudge */
            sscanf(input,"%s %f",string,&polyfudge);
            fprintf(flout,"Fudge factor for polyhedra set to %f. This option should be used with care (!)\n",polyfudge); 
            break;   
        case 27: /* unit cell direction vectors */
            Display_axes = 1;
	    break;
        case 28: /* phong */
            sscanf(input,"%s %f %f",string,&Phong_Value,&Phong_Size);
            break;   
        case 29: /* molcomp */
            sscanf(input,"%s %f",string,&mol_d);
            domolcomp = 1;
            break;   
        case 30: /* cutout */
            El_Cutout = 1;
            sscanf(input,"%s %s",string,Cutout_color);
            break;   
        case 31: /* depthcue - thickness of edges scaled by Z coordinate*/
            sscanf(input,"%s %f",string,&DepthCue);
            break;   
	default: fprintf(flout,"Previous line did not start with a recognizable keyword - ignored\n"); 
	    break;
    }
  }
  if (n_ellips > 1) do_ellipsoids = 1;
  if (packflag != 0) {
    boxlim[0] = boxlim[1] = boxlim[2] = 500.0;
  }
}  /* end of read_inp */

/* ************************************************************** */
/* ************************************************************** */

void Token_Strip(string,no)

char string[100];              /* input output string */
int no;                        /* number of tokens to strip */

/* routine to strip tokens from the leading part of a string */

{
  int i,j;

  for(i=0,j=0;i<no;i++) {  /* loop through tokens to be removed */
    while (string[j++] != ' ') string[j-1] = ' ';  /* characters to space */
    while (string[j] == ' ') j++;                  /* skip all spaces */
  }
  Blank_Strip(string);     /* move stuff to left */
}

/* ************************************************************** */
/* ************************************************************** */

void usage()

{
  fprintf(stderr,"\n****************************************\n\n");
  fprintf(stderr,"Not enough command-line arguments have been specified.\n\n");
  fprintf(stderr,"The full usage is:\n\n");
  fprintf(stderr,"DRAWxtl [-b Xb Yb Zb] [-c] [-C] [-m M] [-n] [-o Xo Yo Zo]\n");
  fprintf(stderr,"        [-p xlo xhi ylo yhi zlo zhi] [-u] [-v Xr Yr Zr] <file>\n\n");
  fprintf(stderr,"-b x-, y-, z-limits of view box in angstroms (20 20 20)\n");
  fprintf(stderr,"-c include comments in the VRML output\n");
  fprintf(stderr,"-C output multiple cameras to the VRML file\n");
  fprintf(stderr,"-m magnification factor for image (1.)\n");
  fprintf(stderr,"-n no display of unit cell\n");
  fprintf(stderr,"-o x-, y-, z-coordinates of origin (0.5 0.5 0.5)\n");
  fprintf(stderr,"-p limits of fract. coord. along a, b, and c in display box (0 1 0 1 0 1)\n");
  fprintf(stderr,"-u display axial orientation vector triple\n");
  fprintf(stderr,"-v xrot yrot zrot : View rotation angles (0. 0. 0.)\n");
  fprintf(stderr,"<file> is the name of the file describing the structure.\n\n");
  fprintf(stderr,"Arguments enclosed in brackets [] are optional.\n");
  fprintf(stderr,"<file> is required.\n\n");
#ifdef _MAC_
  printf("Use Apple Q to close this window.\n");
#endif
  exit(1);
}
